"""Various utilities."""
import random
import string
from functools import partial
from typing import Any, Callable
from unittest.mock import Mock

import grpc
from google.protobuf.message import Message
from google.protobuf.text_format import MessageToString

MESSAGE_MAX_LENGTH = 256


def get_call_id() -> str:
    """Return random call ID for logging."""
    return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(4))  # nosec


def format_message(message: Any) -> str:
    """Format a message to string.

    If argument is a gRPC Message, format it nicely.
    If it's another object, use simple str conversion.
    In any case the output has limited length.
    """
    if isinstance(message, Message):
        out = MessageToString(message, as_one_line=True)
    else:
        out = str(message)
    if len(out) > MESSAGE_MAX_LENGTH:
        return out[:MESSAGE_MAX_LENGTH - 3] + '...'
    else:
        return out


class CallableMock:
    """Mock of a gRPC method Callable object."""

    def __init__(self, method: str, call: Callable):
        self._method = method.encode('utf-8')
        self.call = call

    def __call__(self, *args, **kwargs):
        """Mock gRPC method call."""
        return self.call(*args, **kwargs)

    def future(self, *args, **kwargs):
        """Mock asynchronous gRPC method call."""
        future_mock = Mock(name='future', spec=('result', ))
        future_mock.result.return_value = self.call(*args, **kwargs)
        return future_mock


class MockingChannel(grpc.Channel):
    """A channel which binds Stub calls to TestClient internal mock."""

    def __init__(self, client: 'TestClientMixin'):
        self.client = client

    def subscribe(self, *args, **kwargs):
        """Noop."""

    def unsubscribe(self, *args, **kwargs):
        """Noop."""

    def unary_unary(self, method: str, *args, **kwargs):
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method))

    def unary_stream(self, method: str, *args, **kwargs):
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method))

    def stream_unary(self, method: str, *args, **kwargs):
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method))

    def stream_stream(self, method: str, *args, **kwargs):
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method))

    def close(self, *args, **kwargs):
        """Noop."""


class TestClientMixin:
    """Mixin to wrap gRPC client for tests.

    Attributes:
        mock: A mock which handles all gRPC method calls.
              Acts with a signature `(request: grpc.Message, method: str, timeout: int = None)`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)  # type: ignore
        self.mock = Mock(spec=('__call__', ))

    def connect(self) -> None:
        """Create a mocking channel which redirects all gRPC method calls to the mock attribute."""
        self.channel = MockingChannel(self)
